use crate::{system::{self, Mount, MountType}, result::Result};

trait Platform {
    /// Whether the current Platform is the `Self` platform.
    ///
    /// This method should check for the existence of hardware devices that irrefutably defines
    /// this platform as the given platform. For instance, the existence of `/dev/n`
    fn is(&self) -> Result<bool>;

    /// Return the names and parameters for any required kernel modules.
    fn get_modules(&self) -> Result<Vec<(String, String)>> {
        Ok(vec![])
    }

    /// The configuration for mounting filesystems for the platform.
    ///
    /// This normally includes filesystems such as `/dev` and `/proc` that are not
    /// backed by physical media.
    fn get_mounts(&self) -> Result<Vec<Mount>> {
        use libc::{MS_NODEV, MS_NOEXEC, MS_NOSUID};
        use MountType::{DevPts, DevTmpFs, Proc, Shm, SysFs, TmpFs};

        let no_se = MS_NOSUID | MS_NOEXEC;
        let no_dse = no_se | MS_NODEV;

        let m755 = Some("mode=0755");

        Ok(vec![
            Mount::new(DevTmpFs, "/dev", DevTmpFs, no_se, m755),
            Mount::new(DevPts, "/dev/pts", DevPts, no_se, None),
            Mount::new(Shm, "/dev/shm", TmpFs, no_dse, m755),
            Mount::new(Proc, "/proc", Proc, no_dse, m755),
            Mount::new(TmpFs, "/tmp", TmpFs, no_dse, None),
            Mount::new(SysFs, "/sys", SysFs, no_dse, None),
            Mount::new("cgroup_root", "/sys/fs/cgroup", TmpFs, no_dse, m755),
        ])
    }

    /// Initialize all necessary requirements for the platform.
    fn init(&self) -> Result<()> {
        Ok(())
    }
}

fn init_filesystems(iter: impl IntoIterator<Item = Mount>) -> Result<()> {
    for mount in iter {
        mount.mount_self()?;
    }

    Ok(())
}

fn init_modules(iter: impl IntoIterator<Item = (String, String)>) -> Result<()> {
    for (module, params) in iter {
        system::insmod(&module, &params)?;
    }

    Ok(())
}

#[cfg(feature = "aws")]
mod aws;

fn get_current_platform() -> Result<Option<Box<dyn Platform>>> {
    #[cfg(feature = "aws")]
    if aws::Aws.is()? {
        return Ok(Some(Box::new(aws::Aws)))
    }

    Ok(None)
}

pub fn init() -> Result<()> {
    // Error handling strategy: If a platform is compiled in and loaded, if platform
    // specific error handling doesn't work, fall back to generic.

    let platform = get_current_platform()?;
    if let Some(platform) = platform {
        // NOTE: We need to make get_mounts _additional_ beyond a base set.
        // We need `/dev/nsm` to exist so Aws.is() works.
        platform.get_mounts().and_then(init_filesystems)?;
        platform.get_modules().and_then(init_modules)?;
        platform.init()?;
    }

    Ok(())
}
