use crate::result::{Result, Context};
use std::path::{PathBuf, Path};

pub mod syscall;

pub enum MountType {
    DevTmpFs,
    DevPts,
    Shm,
    Proc,
    TmpFs,
    SysFs,
}

impl From<MountType> for PathBuf {
    fn from(value: MountType) -> Self {
        Self::from(value.as_str())
    }
}

impl MountType {
    fn as_str(&self) -> &'static str {
        match self {
            MountType::DevTmpFs => "devtmpfs",
            MountType::DevPts => "devpts",
            MountType::Shm => "shm",
            MountType::Proc => "proc",
            MountType::TmpFs => "tmpfs",
            MountType::SysFs => "sysfs",
        }
    }
}

pub struct Mount {
    source: PathBuf,
    target: PathBuf,
    fstype: MountType,
    flags: libc::c_ulong,
    data: Option<String>,
}

impl Mount {
    pub fn new<'a>(
        source: impl Into<PathBuf>,
        target: impl Into<PathBuf>,
        fstype: MountType,
        flags: libc::c_ulong,
        data: impl Into<Option<&'a str>>,
    ) -> Self {
        let source = source.into();
        let target = target.into();
        let data = data.into().map(Into::into);

        Self {
            source,
            target,
            fstype,
            flags,
            data,
        }
    }

    pub fn mount_self(&self) -> Result<()> {
        syscall::mount(
            &self.source,
            &self.target,
            self.fstype.as_str(),
            self.flags,
            self.data.as_deref(),
        )?;
        Ok(())
    }
}

pub fn insmod(path: impl AsRef<Path>, params: impl AsRef<str>) -> Result<()> {
    let path = path.as_ref();
    let params = params.as_ref();
    let file = std::fs::File::open(path).context(format_args!(
        "could not open: {path}",
        path = path.display()
    ))?;

    syscall::finit_module(file, params)?;
    Ok(())
}
